from typing import Any  # noqa
from typing import Dict, List, Optional

from pydantic import BaseModel, Field


class Command(BaseModel):
    name: "str" = Field(..., alias="name")
    description: "Optional[str]" = Field(None, alias="description")
    arguments: "Optional[List[CommandArgument]]" = Field(None, alias="arguments")


class CommandArgument(BaseModel):
    name: "str" = Field(..., alias="name")
    is_optional: "Optional[bool]" = Field(None, alias="is_optional")
    description: "Optional[str]" = Field(None, alias="description")


class GetModulesResponse(BaseModel):
    modules: "Dict[str, Module]" = Field(..., alias="modules")


class HTTPValidationError(BaseModel):
    detail: "Optional[List[ValidationError]]" = Field(None, alias="detail")


class Message(BaseModel):
    channel_id: "str" = Field(..., alias="channel_id")
    text: "str" = Field(..., alias="text")
    blocks: "Optional[List[Any]]" = Field(None, alias="blocks")
    send_ephemeral: "Optional[bool]" = Field(None, alias="send_ephemeral")
    user_id: "Optional[str]" = Field(None, alias="user_id")


class Module(BaseModel):
    name: "str" = Field(..., alias="name")
    description: "Optional[str]" = Field(None, alias="description")
    url: "str" = Field(..., alias="url")
    commands: "Dict[str, Command]" = Field(..., alias="commands")


class SendMessageResponse(BaseModel):
    ok: "bool" = Field(..., alias="ok")


class ValidationError(BaseModel):
    loc: "List[str]" = Field(..., alias="loc")
    msg: "str" = Field(..., alias="msg")
    type: "str" = Field(..., alias="type")



import asyncio
import logging
from asyncio import iscoroutinefunction, Task
from dataclasses import dataclass, field
from inspect import signature, Parameter
from typing import Callable, Optional, Dict, List, Any, Type

from fastapi import FastAPI, Depends

from fastapi_metabot.client import (
    AsyncApis,
    ApiClient,
    models
)
from fastapi_metabot.client.exceptions import ApiException

log = logging.getLogger(__name__)

Converter = Callable[[str], Any]


@dataclass
class Argument:
    name: str
    is_optional: bool
    type: Type = str  # noqa A003
    description: Optional[str] = None


@dataclass
class Command:
    name: str
    func: Callable
    description: Optional[str] = None
    arguments: List[Argument] = field(default_factory=list)


class Module:
    name: str
    description: Optional[str]
    module_url: str
    metabot_client: ApiClient
    heartbeat_delay: float

    _commands: Dict[str, Command]
    _converters: Dict[Type, Converter]
    _heartbeat: Optional[Task]

    def __init__(
            self,
            name: str,
            module_url: str,
            metabot_url: str,
            description: Optional[str] = None,
            heartbeat_delay: float = 10,
    ) -> None:
        self.name = name
        self.description = description
        self.module_url = module_url
        self.metabot_client = ApiClient(host=metabot_url)
        self.heartbeat_delay = heartbeat_delay

        self._commands = {}
        self._converters = {}
        self._heartbeat = None

    def command(
            self,
            name: str,
            *,
            function: Optional[Callable] = None,
            description: Optional[str] = None,
            arg_descriptions: Optional[Dict[str, str]] = None,
    ) -> Callable:
        def wrapper(f: Callable) -> Callable:
            arguments = self._parse_arguments(f, arg_descriptions)
            self._commands[name] = Command(
                name=name,
                func=f,
                description=description,
                arguments=arguments
            )
            return f

        if function is None:
            return wrapper
        else:
            return wrapper(function)

    def converter(
            self,
            to_type: Type,
            *,
            converter: Optional[Converter] = None,
    ) -> Callable:
        def wrapper(f: Callable) -> Callable:
            self._converters[to_type] = f
            return f

        if converter is None:
            return wrapper
        else:
            return wrapper(converter)

    @staticmethod
    def _parse_arguments(
            f: Callable,
            arg_descriptions: Optional[Dict[str, str]] = None,
    ) -> List[Argument]:
        sig = signature(f)
        params = sig.parameters

        allowed_kind = (Parameter.POSITIONAL_OR_KEYWORD, Parameter.KEYWORD_ONLY)
        assert all(param.kind in allowed_kind for param in params.values()), (
            'Positional-only arguments, *args and **kwargs are not allowed'
        )

        arg_descriptions = arg_descriptions or {}
        return [Argument(
            name=p.name,
            is_optional=p.default is not Parameter.empty,
            type=p.annotation if p.annotation is not Parameter.empty else str,
            description=arg_descriptions.get(p.name),
        ) for p in params.values()]

    async def execute_command(
            self,
            name: str,
            arguments: Dict[str, str],
    ) -> None:
        command = self._commands[name]
        converted_args = {
            arg.name: self._get_converter(arg.type)(arguments[arg.name])
            for arg
            in command.arguments
        }
        if iscoroutinefunction(command.func):
            await command.func(**converted_args)
        else:
            command.func(**converted_args)

    def _get_converter(self, to_type: Type) -> Converter:
        return self._converters.get(to_type, to_type)

    def install(self, app: FastAPI, prefix: str = '') -> FastAPI:
        from fastapi_metabot.routes import router
        from fastapi_metabot.utils import current_module

        def set_current_module() -> None:
            current_module.set(self)

        app.include_router(
            router,
            prefix=prefix,
            tags=['metabot', self.name],
            dependencies=[Depends(set_current_module)],
        )

        app.add_event_handler('startup', self._start_heartbeat)
        app.add_event_handler('shutdown', self._stop_heartbeat)

        return app

    def _start_heartbeat(self) -> None:
        module = models.Module(
            name=self.name,
            description=self.description,
            url=self.module_url,
            commands={command.name: models.Command(
                name=command.name,
                description=command.description,
                arguments=[models.CommandArgument(
                    name=arg.name,
                    description=arg.description,
                    is_optional=arg.is_optional,
                ) for arg in command.arguments],
            ) for command in self._commands.values()},
        )
        async_api = AsyncApis(self.metabot_client)

        async def heartbeat() -> None:
            while True:
                try:
                    await (
                        async_api.metabot_api.
                        register_module_api_modules_post(module)
                    )
                    log.debug('Heartbeat to metabot server was successful')
                except ApiException:
                    log.exception('Heartbeat to metabot server has failed')

                await asyncio.sleep(self.heartbeat_delay)

        self._heartbeat = asyncio.create_task(heartbeat())

    def _stop_heartbeat(self) -> None:
        if self._heartbeat is not None:
            self._heartbeat.cancel()



# flake8: noqa E501
from asyncio import get_event_loop
from typing import TYPE_CHECKING, Awaitable

from fastapi.encoders import jsonable_encoder

from fastapi_metabot.client import models as m

if TYPE_CHECKING:
    from fastapi_metabot.client.api_client import ApiClient


class _MetabotApi:
    def __init__(self, api_client: "ApiClient"):
        self.api_client = api_client

    def _build_for_get_module_by_name_api_modules_module_name_get(self, module_name: str) -> Awaitable[m.Module]:
        path_params = {"module_name": str(module_name)}

        return self.api_client.request(
            type_=m.Module, method="GET", url="/api/modules/{module_name}", path_params=path_params,
        )

    def _build_for_get_modules_api_modules_get(self,) -> Awaitable[m.GetModulesResponse]:
        return self.api_client.request(type_=m.GetModulesResponse, method="GET", url="/api/modules/",)

    def _build_for_register_module_api_modules_post(self, module: m.Module) -> Awaitable[m.Module]:
        body = jsonable_encoder(module)

        return self.api_client.request(type_=m.Module, method="POST", url="/api/modules/", json=body)

    def _build_for_send_message_to_slack_api_chat_post(self, message: m.Message) -> Awaitable[m.SendMessageResponse]:
        body = jsonable_encoder(message)

        return self.api_client.request(type_=m.SendMessageResponse, method="POST", url="/api/chat/", json=body)


class AsyncMetabotApi(_MetabotApi):
    async def get_module_by_name_api_modules_module_name_get(self, module_name: str) -> m.Module:
        return await self._build_for_get_module_by_name_api_modules_module_name_get(module_name=module_name)

    async def get_modules_api_modules_get(self,) -> m.GetModulesResponse:
        return await self._build_for_get_modules_api_modules_get()

    async def register_module_api_modules_post(self, module: m.Module) -> m.Module:
        return await self._build_for_register_module_api_modules_post(module=module)

    async def send_message_to_slack_api_chat_post(self, message: m.Message) -> m.SendMessageResponse:
        return await self._build_for_send_message_to_slack_api_chat_post(message=message)


class SyncMetabotApi(_MetabotApi):
    def get_module_by_name_api_modules_module_name_get(self, module_name: str) -> m.Module:
        coroutine = self._build_for_get_module_by_name_api_modules_module_name_get(module_name=module_name)
        return get_event_loop().run_until_complete(coroutine)

    def get_modules_api_modules_get(self,) -> m.GetModulesResponse:
        coroutine = self._build_for_get_modules_api_modules_get()
        return get_event_loop().run_until_complete(coroutine)

    def register_module_api_modules_post(self, module: m.Module) -> m.Module:
        coroutine = self._build_for_register_module_api_modules_post(module=module)
        return get_event_loop().run_until_complete(coroutine)

    def send_message_to_slack_api_chat_post(self, message: m.Message) -> m.SendMessageResponse:
        coroutine = self._build_for_send_message_to_slack_api_chat_post(message=message)
        return get_event_loop().run_until_complete(coroutine)



from contextvars import ContextVar
from typing import Optional

from fastapi_metabot.models import SlackMetadata
from fastapi_metabot.module import Module


slack_metadata: ContextVar[Optional[SlackMetadata]] = ContextVar(
    'slack_metadata',
    default=None
)

current_module: ContextVar[Optional[Module]] = ContextVar(
    'current_module',
    default=None
)



import json
from typing import Any, Dict, Optional

from httpx import Response, Headers

MAX_CONTENT = 200


class ApiException(Exception):
    """Base class"""


class UnexpectedResponse(ApiException):
    def __init__(self, status_code: Optional[int], reason_phrase: str, content: bytes, headers: Headers) -> None:
        self.status_code = status_code
        self.reason_phrase = reason_phrase
        self.content = content
        self.headers = headers

    @staticmethod
    def for_response(response: Response) -> "ApiException":
        return UnexpectedResponse(
            status_code=response.status_code,
            reason_phrase=response.reason_phrase,
            content=response.content,
            headers=response.headers,
        )

    def __str__(self) -> str:
        status_code_str = f"{self.status_code}" if self.status_code is not None else ""
        if self.reason_phrase == "" and self.status_code is not None:
            reason_phrase_str = "(Unrecognized Status Code)"
        else:
            reason_phrase_str = f"({self.reason_phrase})"
        status_str = f"{status_code_str} {reason_phrase_str}".strip()
        short_content = self.content if len(self.content) <= MAX_CONTENT else self.content[: MAX_CONTENT - 3] + b" ..."
        raw_content_str = f"Raw response content:\n{short_content!r}"
        return f"Unexpected Response: {status_str}\n{raw_content_str}"

    def structured(self) -> Dict[str, Any]:
        return json.loads(self.content)


class ResponseHandlingException(ApiException):
    def __init__(self, source: Exception):
        self.source = source



from fastapi import BackgroundTasks, APIRouter, HTTPException
from fastapi.encoders import jsonable_encoder

from fastapi_metabot.models import MetabotPayload
from fastapi_metabot.utils import current_module, slack_metadata

router = APIRouter()


@router.post('/commands/{command_name}')
async def execute_command(
        command_name: str,
        payload: MetabotPayload,
        background_tasks: BackgroundTasks,
) -> None:
    module = current_module.get()
    if module is None:
        raise HTTPException(500)

    slack_metadata.set(payload.metadata)

    background_tasks.add_task(
        module.execute_command,
        command_name,
        jsonable_encoder(payload.arguments),
    )



from asyncio import get_event_loop
from functools import lru_cache
from typing import Any, Awaitable, Callable, Dict, Generic, Type, TypeVar, overload

from httpx import AsyncClient, Request, Response
from pydantic import ValidationError

from fastapi_metabot.client.api.metabot_api import AsyncMetabotApi, SyncMetabotApi
from fastapi_metabot.client.exceptions import ResponseHandlingException, UnexpectedResponse

ClientT = TypeVar("ClientT", bound="ApiClient")


class AsyncApis(Generic[ClientT]):
    def __init__(self, client: ClientT):
        self.client = client

        self.metabot_api = AsyncMetabotApi(self.client)


class SyncApis(Generic[ClientT]):
    def __init__(self, client: ClientT):
        self.client = client

        self.metabot_api = SyncMetabotApi(self.client)


T = TypeVar("T")
Send = Callable[[Request], Awaitable[Response]]
MiddlewareT = Callable[[Request, Send], Awaitable[Response]]


class ApiClient:
    def __init__(self, host: str = None, **kwargs: Any) -> None:
        self.host = host
        self.middleware: MiddlewareT = BaseMiddleware()
        self._async_client = AsyncClient(**kwargs)

    @overload
    async def request(
        self, *, type_: Type[T], method: str, url: str, path_params: Dict[str, Any] = None, **kwargs: Any
    ) -> T:
        ...

    @overload  # noqa F811
    async def request(
        self, *, type_: None, method: str, url: str, path_params: Dict[str, Any] = None, **kwargs: Any
    ) -> None:
        ...

    async def request(  # noqa F811
        self, *, type_: Any, method: str, url: str, path_params: Dict[str, Any] = None, **kwargs: Any
    ) -> Any:
        if path_params is None:
            path_params = {}
        url = (self.host or "") + url.format(**path_params)
        request = Request(method, url, **kwargs)
        return await self.send(request, type_)

    @overload
    def request_sync(self, *, type_: Type[T], **kwargs: Any) -> T:
        ...

    @overload  # noqa F811
    def request_sync(self, *, type_: None, **kwargs: Any) -> None:
        ...

    def request_sync(self, *, type_: Any, **kwargs: Any) -> Any:  # noqa F811
        """
        This method is not used by the generated apis, but is included for convenience
        """
        return get_event_loop().run_until_complete(self.request(type_=type_, **kwargs))

    async def send(self, request: Request, type_: Type[T]) -> T:
        response = await self.middleware(request, self.send_inner)
        if response.status_code in [200, 201]:
            try:
                return parse_as_type(response.json(), type_)
            except ValidationError as e:
                raise ResponseHandlingException(e)
        raise UnexpectedResponse.for_response(response)

    async def send_inner(self, request: Request) -> Response:
        try:
            response = await self._async_client.send(request)
        except Exception as e:
            raise ResponseHandlingException(e)
        return response

    def add_middleware(self, middleware: MiddlewareT) -> None:
        current_middleware = self.middleware

        async def new_middleware(request: Request, call_next: Send) -> Response:
            async def inner_send(request: Request) -> Response:
                return await current_middleware(request, call_next)

            return await middleware(request, inner_send)

        self.middleware = new_middleware


class BaseMiddleware:
    async def __call__(self, request: Request, call_next: Send) -> Response:
        return await call_next(request)


@lru_cache(maxsize=None)
def _get_parsing_type(type_: Any, source: str) -> Any:
    from pydantic.main import create_model

    type_name = getattr(type_, "__name__", str(type_))
    return create_model(f"ParsingModel[{type_name}] (for {source})", obj=(type_, ...))


def parse_as_type(obj: Any, type_: Type[T]) -> T:
    model_type = _get_parsing_type(type_, source=parse_as_type.__name__)
    return model_type(obj=obj).obj



from typing import Dict

from pydantic import BaseModel


class SlackMetadata(BaseModel):
    token: str
    command: str
    response_url: str
    trigger_id: str
    user_id: str
    user_name: str
    channel_id: str
    text: str


class MetabotPayload(BaseModel):
    arguments: Dict[str, str]
    metadata: SlackMetadata



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
import os
from pathlib import Path

from setuptools import find_packages, setup


# Package meta-data.
NAME = 'fastapi-metabot'
DESCRIPTION = 'FastAPI plugin for quickly developing MetaBot modules'
URL = 'https://github.com/vasyukvv42/metabot'
EMAIL = 'vasyukvv42@gmail.com'
AUTHOR = 'Vladislav Vasyuk'
REQUIRES_PYTHON = '>=3.8.0'


# What packages are required for this module to be executed?
def list_reqs(fname='requirements.txt'):
    with open(fname) as fd:
        return [x for x in fd.read().splitlines() if not x.startswith('#')]


# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the
# Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION


# Load the package's __version__.py module as a dictionary.
ROOT_DIR = Path(__file__).resolve().parent
PACKAGE_DIR = ROOT_DIR / NAME.replace('-', '_')
about = {}
with open(PACKAGE_DIR / 'VERSION') as f:
    _version = f.read().strip()
    about['__version__'] = _version


# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=('tests',)),
    package_data={'fastapi_metabot': ['VERSION']},
    install_requires=list_reqs(),
    extras_require={},
    include_package_data=True,
    license='MIT',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
)



import inspect

from fastapi_metabot.client import models
from fastapi_metabot.client.api_client import ApiClient, AsyncApis, SyncApis  # noqa F401

for model in inspect.getmembers(models, inspect.isclass):
    if model[1].__module__ == "fastapi_metabot.client.models":
        model_class = model[1]
        model_class.update_forward_refs()
